/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * annunciators.c
 *
 * Version 1.0
 *
 * Created: 29.05.2017
 * Author: Mario Hellmich <mario.hellmich@web.de>
 *
 * Annunciator multiplexing and output
 */ 


#include <avr/io.h>
#include <avr/interrupt.h>

#include "annunciators.h"
#include "def.h"

extern volatile uint32_t instr_state;
extern volatile uint8_t  instr_error;
extern volatile uint8_t  annunc_pattern[13];
extern volatile uint32_t tuned_freq;
extern volatile int16_t  bfo_offset;
extern volatile uint8_t  cor_level;
extern volatile uint8_t  rssi_reading;




/*
************************************************************************************************************************************************
Display and Annunciator Multiplexing:
Multiplexing is performed digit-wise in the seven segment displays and is triggered by Timer0. The pushbutton annunciators and indicator
LEDs are combined in three groups of eight LEDs (the last group is partially empty), the groups are illuminated consecutively after the
seven segment displays. Multiplexing clock is 1ms.
*/

static volatile uint8_t mux_count = 0;

#define ANNUNC_PORT				PORTA
#define ANNUNC_DDR				DDRA
#define MUX_PORT				PORTB
#define MUX_DDR					DDRB
#define MUX_OUT0				PB4
#define MUX_OUT1				PB5
#define MUX_OUT2				PB6
#define MUX_OUT3				PB7
#define ANNUNC_DEACTIVATE		PORTA = 0x00


void annunc_init(void)
{
	ANNUNC_DDR = 0xFF;
	MUX_DDR   |= (1 << MUX_OUT0) | (1<<MUX_OUT1) | (1<<MUX_OUT2) | (1<<MUX_OUT3);
	ANNUNC_DEACTIVATE;
	ANNUNC_PORT &= ~( (1 << MUX_OUT0) | (1<<MUX_OUT1) | (1<<MUX_OUT2) | (1<<MUX_OUT3) );
	TCCR0 |= (1<<CS00) | (1<<CS02);															// Timer0 prescaler set to divide by 128
	TCCR0 |= (1<<WGM01);																	// Clear timer on compare
	TIMSK |= (1<<OCIE0);																	// Timer0 output compare match interrupt enable
	OCR0 = 125;																				// Compare Register, set to 125, interrupt approx every 5ms
}

ISR(TIMER0_COMP_vect)																		// Interrupt service routine for Timer0 compare match
{
	ANNUNC_DEACTIVATE;
	NOP();
	MUX_PORT = (MUX_PORT & 0x0F) | (mux_count << 4);
	NOP();
	ANNUNC_PORT = annunc_pattern[mux_count];
	if(mux_count < 12)
		++mux_count;
	else
		mux_count = 0;
}




/*
************************************************************************************************************************************************
Data Output to Seven Segment Displays and Annunciators
*/

// Seven segment display patterns
#define DISP_0					0b00111111    // 0: a, b, c, d, e, f
#define DISP_1					0b00000110    // 1: b, c
#define DISP_2					0b01011011    // 2: a, b, d, e, g
#define DISP_3					0b01001111    // 3: a, b, c, d, g
#define DISP_4					0b01100110    // 4: b, c, f, g
#define DISP_5					0b01101101    // 5: a, c, d, f, g
#define DISP_6					0b01111101    // 6: a, c, d, e, f, g
#define DISP_7					0b00000111    // 7: a, b, c
#define DISP_8					0b01111111    // 8: a, b, c, d, e, f, g
#define DISP_9					0b01101111    // 9: a, b, c, d, f, g
#define DISP_BLANK				0b00000000	  // Blank
#define DISP_DP					0b10000000	  // Decimal Point
#define DISP_MINUS				0b01000000	  // Minus or Bar

static uint8_t segments (uint8_t digit)
{
	switch(digit)
	{
		case 0:
			return DISP_0;
			break;
		case 1:
			return DISP_1;
			break;
		case 2:
			return DISP_2;
			break;
		case 3:
			return DISP_3;
			break;
		case 4:
			return DISP_4;
			break;
		case 5:
			return DISP_5;
			break;
		case 6:
			return DISP_6;
			break;
		case 7:
			return DISP_7;
			break;
		case 8:
			return DISP_8;
			break;
		case 9:
			return DISP_9;
			break;
		default:
			return DISP_BLANK;
			break;
	}
}

static void display_freq(uint32_t freq)
{
	uint8_t i;
	for ( i = 7; i != 0; --i )																// Update frequency display
	{
		if( !( (i-1 == 0) && (freq == 0) ) )												// Suppress leading zeros: blank digit if freq is zero on MSD
		{
			cli();
			annunc_pattern[i-1] = segments(freq % 10);										// Write digit to annunc_pattern, global interrupts off until digit is written
			if( i - 1 == 1 )
				annunc_pattern[i-1] |= DISP_DP;												// Set decimal point on MSD-1
			sei();
		}
		else
			annunc_pattern[i-1] = DISP_BLANK;
		freq = freq / 10;
	}
}

static void display_bfo_offs(int16_t bfo_offs)
{
	uint8_t i;
	uint8_t sign;
	if( bfo_offs < 0 )																		// BFO offset is negative?
	{
		bfo_offs = -bfo_offs;																// Then change sign
		sign = 1;																			// Remember sign
	}
	else
		sign = 0;
	for ( i = 6; i != 2; --i )																// Update BFO display, BFO offset is displayed on the four LSD
	{
		cli();
		annunc_pattern[i] = segments(bfo_offs % 10);										// Write digit to annunc_pattern, global interrupts off until digit is written
		if( i == 3 )
			annunc_pattern[i] |= DISP_DP;													// Set decimal point on MSD-1 (reading is in kHz)
		bfo_offs = bfo_offs / 10;
		sei();
	}
	annunc_pattern[0]  = DISP_BLANK;
	annunc_pattern[1]  = DISP_BLANK;														// MSD and MSD-1 are blank
	if( sign == 0 )
		annunc_pattern[2] = DISP_BLANK;														// Display no sign if BFO tuning offset is nonnegative
	else
		annunc_pattern[2] = DISP_MINUS;														// Display minus if BFO tuning offset is negative
}

static void display_rssi_reading(uint8_t rssi_reading)
{
	uint8_t i;
	for ( i = 3; i != 0; --i )																// Display RSSI reading
	{
		if( !( (i+6 <= 8) && (rssi_reading == 0) ) )										// Suppress leading zeros: blank digit if rssi_reading is zero on the two MSD
			annunc_pattern[i+6] = segments(rssi_reading % 10);
		else
			annunc_pattern[i+6] = DISP_BLANK;
		rssi_reading = rssi_reading / 10;
	}
}

static void display_cor_level(uint8_t cor_level)
{
	uint8_t i;
	if ( cor_level >= MIN_COR_LEVEL )														// COR level below minimum?
	{
		annunc_pattern[7] = DISP_MINUS;														// Display "---"
		annunc_pattern[8] = DISP_MINUS;
		annunc_pattern[9] = DISP_MINUS;
	}
	else																					// COR level is above minimum
	{
		for ( i = 3; i != 0; --i )															// Display COR level
		{
			annunc_pattern[i+6] = segments(cor_level % 10);
			cor_level = cor_level / 10;
		}
	}
}

static void display_error(uint8_t error_disp)
{
	annunc_pattern [0]=  0b01111001;
	annunc_pattern [1]=  0b01010000;
	annunc_pattern [2]=  0b01010000;
	annunc_pattern [3]=  0x00;																// Display "Err_" on frequency display

	annunc_pattern [7]=  0x00;
	annunc_pattern [8]=  0x00;
	annunc_pattern [9]=  0x00;																// Blank RSSI display	
	
	uint8_t i;
	for( i = 7; i != 4; i-- )																// Display error code on frequency display
	{
		annunc_pattern[i-1] = segments(error_disp % 10);
		error_disp = error_disp / 10;
	}
}

void disp_out(void)																			// Write output to the seven segment registers, to be called periodically in main function
{
	static uint32_t freq_old;
	static int16_t bfo_offs_old;
	static uint8_t rssi_reading_old;
	static uint8_t cor_levl_old;	
	static uint32_t instr_state_old;
		
	cli();
	uint32_t freq_loc = tuned_freq;															// Make local copies, reactivate global interrupts in between to avoid losing interrupts
	int16_t  bfo_offs_loc = bfo_offset;
	sei(); cli();
	uint8_t rssi_reading_loc = rssi_reading;
	uint8_t cor_level_loc = cor_level;
	uint32_t instr_state_loc = instr_state;
	sei();
	
	if( instr_error == 0 )																	// No error flag is set? If an error flag is set, display output is handled by display_error in fault.c
	{
		if( (freq_loc != freq_old) || (bfo_offs_loc != bfo_offs_old) || (instr_state_loc != instr_state_old) )				// Frequency, BFO offset or instr_state has changed?
		{
			freq_old = freq_loc;
			bfo_offs_old = bfo_offs_loc;
			if( !(instr_state & (1UL << BFO_TUNE)) )										// Receiver is not in BFO tune mode?
				display_freq(freq_loc / 10);												// Divide by 10 since display and tuning resolution is 10 Hz
			else																			// Receiver is in BFO tune mode
				display_bfo_offs(bfo_offs_loc);
		}
		if( (rssi_reading_loc != rssi_reading_old) || (cor_level_loc != cor_levl_old) || (instr_state_loc != instr_state_old) )	// rssi_reading, cor_level or instr_state has changed?
		{
			rssi_reading_old = rssi_reading_loc;
			cor_levl_old = cor_level_loc;
			if( !(instr_state & (1UL << SET_COR)) )											// Receiver is not in SET COR level mode?
				display_rssi_reading(rssi_reading_loc);
			else																			// Receiver is in SET COR level mode
				display_cor_level(cor_level_loc);
		}
	}
	else
		display_error(instr_error);															// Display instr_error as decimal value on RSSI display	
}

void annunc_out(void)																		// Write output to the annunciator registers, to be called periodically in main function
{
	static uint32_t instr_state_old;
	cli();
	uint32_t instr_state_loc = instr_state;													// Make local copy of instr_state
	sei();
	
	if( instr_state_loc != instr_state_old )												// instr_state has changed?
	{
		instr_state_old = instr_state_loc;
		annunc_pattern[10] = (uint8_t)
		(
		  ( ( (instr_state_loc >> AM)             & 0x01 ) << 0 )
		| ( ( (instr_state_loc >> USB)            & 0x01 ) << 1 )
		| ( ( (instr_state_loc >> LSB)            & 0x01 ) << 2 )
		| ( ( (instr_state_loc >> MUTE_SELECTED)  & 0x01 ) << 3 )
		| ( ( (instr_state_loc >> AGC_FAST)       & 0x01 ) << 4 )
		| ( ( (instr_state_loc >> AGC_SLOW)       & 0x01 ) << 5 )
		| ( ( (instr_state_loc >> AGC_HANG)       & 0x01 ) << 6 )
		| ( ( (instr_state_loc >> MGC)            & 0x01 ) << 7 )
		);
		annunc_pattern[11] = (uint8_t)
		(
	   	  ( ( (instr_state_loc >> BW1)            & 0x01 ) << 0 )
		| ( ( (instr_state_loc >> BW2)            & 0x01 ) << 1 )
		| ( ( (instr_state_loc >> BW3)            & 0x01 ) << 2 )
		| ( ( (instr_state_loc >> TUNE_LOCK)      & 0x01 ) << 3 )
		| ( ( (instr_state_loc >> TUNE_10HZ)      & 0x01 ) << 4 )
		| ( ( (instr_state_loc >> TUNE_100HZ)     & 0x01 ) << 5 )
		| ( ( (instr_state_loc >> TUNE_1KHZ)      & 0x01 ) << 6 )
		| ( ( (instr_state_loc >> TUNE_10KHZ)     & 0x01 ) << 7 )
		);
		annunc_pattern[12] = (uint8_t)
		(
		  ( ( (instr_state_loc >> TUNE_100KHZ)    & 0x01 ) << 0 )
		| ( ( (instr_state_loc >> SET_COR)        & 0x01 ) << 1 )
		| ( ( (instr_state_loc >> BFO_TUNE)       & 0x01 ) << 2 )
		| ( ( (instr_state_loc >> ATT_ON)         & 0x01 ) << 3 )
		| ( ( (instr_state_loc >> COR_ACTIVE)     & 0x01 ) << 4 )
		);
	}
}