/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * encoder.c
 *
 * Version 1.0
 *
 * Created: 29.05.2017
 * Author: Mario Hellmich <mario.hellmich@web.de>
 *
 * Rotary encoder poll and updating of appropriate globals
 */ 


#include <avr/io.h>
#include <avr/interrupt.h>

#include "encoder.h"
#include "def.h"

extern volatile uint32_t instr_state;
extern volatile uint32_t tuned_freq;
extern volatile int16_t  bfo_offset;
extern volatile uint8_t  cor_level;




/*
************************************************************************************************************************************************
Rotary Encoder:
Reading the rotary encoder is performed every 2ms.
Credits: Peter Dannegger, http://www.mikrocontroller.net/articles/drehgeber
*/

#define ENC_PORT			PORTB
#define ENC_PIN				PINB
#define ENC_DDR				DDRB
#define ENC_A				PB2
#define ENC_B				PB3
#define ENC_PINA			PINB2
#define ENC_PINB			PINB3

volatile int8_t enc_delta;
static int8_t enc_last;

void enc_init(void)
{
	TCCR1B |= (1<< WGM12);																	// Clear Timer1 on compare
	TCCR1B |= (1<< CS11)|(1<<CS10);															// Set Timer1 prescaler to divide by 64
	TIMSK  |= (1<<OCIE1A);																	// Timer1 output compare match A interrupt enable
	OCR1AH = 0x00;
	OCR1AL = 24;																			// Compare register, set to 24 = 25-1, interrupt every 100s
	ENC_DDR &= ~( (1<<ENC_A) | (1<<ENC_B) );												// Set ENC_A and ENC_B as inputs
	ENC_PIN &= ~( (1<<ENC_A) | (1<<ENC_B) );												// Pull-up off
	int8_t enc_new;
	if( ENC_PIN & (1<<ENC_PINA) )
		enc_new = 3;
	if( ENC_PIN & (1<<ENC_PINB) )
		enc_new ^= 1;
	enc_last = enc_new;																		// Power-on state
	enc_delta = 0;																			// Power-on state
}

ISR(TIMER1_COMPA_vect)																		// ISR for Timer1 compare match A
{
	int8_t enc_new, diff;
	enc_new = 0;
	if( ENC_PIN & (1<<ENC_PINA) )
		enc_new = 3;
	if( ENC_PIN & (1<<ENC_PINB) )
		enc_new ^= 1;
	diff = enc_last - enc_new;
	if( diff & 1 )
	{
		enc_last = enc_new;
		enc_delta += (diff & 2) - 1;
	}
}

int8_t get_enc(void)																		// Returns the number of steps since last call
{
	int8_t retval;
	cli();
	retval = enc_delta;
	enc_delta = retval & 1;
	sei();
	return retval >> 1;
}

uint32_t get_multiplier(void)																// Determines appropriate multiplier to update frequency depending on selected tuning speed
{
	uint32_t multiplier = 0;
	if( instr_state & (1UL<<TUNE_100KHZ) )
		multiplier = 100000;
	else if( instr_state & (1UL<<TUNE_10KHZ) )
		multiplier = 10000;
	else if( instr_state & (1UL<<TUNE_1KHZ) )
		multiplier = 1000;
	else if( instr_state & (1UL<<TUNE_100HZ) )
		multiplier = 100;
	else if( instr_state & (1UL<<TUNE_10HZ) )
		multiplier = 10;
	return multiplier;
}

void read_enc(void)																			// Reads the encoder and updates appropriate globals
{
	int8_t new_delta = -get_enc();
	if( new_delta )																			// Encoder steps nonzero since last call? Only then the globals need modification
	{
		if( !( instr_state & ( (1UL<<BFO_TUNE) | (1UL<<SET_COR) | (1UL<<TUNE_LOCK) ) ) )	// Is receiver in frequency tune mode and tuning not locked?
		{
			int32_t tuning_step = get_multiplier() * new_delta; 
			if( tuning_step > 0 )
			{
				if( tuned_freq + tuning_step <= MAX_FREQ )
					tuned_freq += tuning_step;
				else
					tuned_freq = MAX_FREQ;
			}
			else
			{
				if( ( ( tuned_freq >= -tuning_step ) && (tuned_freq + tuning_step) >= MIN_FREQ ) )
					tuned_freq += tuning_step;
				else
					tuned_freq = MIN_FREQ;
			}
		}
		else if( instr_state & (1UL << BFO_TUNE) )												// Is receiver in BFO tune mode?
		{
			if( ( (bfo_offset + new_delta) <= MAX_BFO_OFFS ) && ( (bfo_offset + new_delta) >= MIN_BFO_OFFS ) )
				bfo_offset += new_delta;
			else if( bfo_offset + new_delta > MAX_BFO_OFFS )
				bfo_offset = MAX_BFO_OFFS;
			else if( bfo_offset + new_delta < MIN_BFO_OFFS )
				bfo_offset = MIN_BFO_OFFS;
		}
		else if( instr_state & (1UL << SET_COR) )												// Is receiver in COR level set mode?
		{
			if( new_delta > 0)
			{
				if( cor_level + new_delta <= MIN_COR_LEVEL )
					cor_level += new_delta;
				else
					cor_level = MIN_COR_LEVEL;
			}
			else
			{
				if( ( cor_level >= -new_delta ) && ( cor_level + new_delta >= MAX_COR_LEVEL ) )
					cor_level += new_delta;
				else
					cor_level = MAX_COR_LEVEL;
			}
		}
	}
}

