/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * instr_control.c
 *
 * Version 1.0
 *
 * Created: 29.05.2017
 * Author: Mario Hellmich <mario.hellmich@web.de>
 *
 * Data output to the MCU ports
 */ 


#include "def.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "instr_control.h"

extern volatile uint32_t instr_state;
extern volatile uint8_t  instr_error;
extern volatile uint8_t  cor_level;
extern volatile uint8_t  rssi_reading;




/*
************************************************************************************************************************************************
Instrument Control Output:
These routines output the instrument state to the appropriate MC ports.
*/

#define CONTR_PORT1			PORTE
#define CONTR_PORT2			PORTC
#define CONTR_PORT3			PORTB
#define CONTR_PORT_AUX		PORTG
#define CONTR_DDR1			DDRE
#define CONTR_DDR2			DDRC
#define CONTR_DDR3			DDRB
#define CONTR_DDR_AUX		DDRG
#define PIN_DET1			PE2
#define PIN_DET2			PE3
#define PIN_ATT				PE4
#define PIN_AGC_ON			PE5
#define PIN_AGC_LONG		PE6
#define PIN_AGC_SHORT		PE7
#define PIN_BWC1			PC5
#define PIN_BWC2			PC6
#define PIN_COR				PB0
#define PIN_STATUSLED		PG0

void contr_init(void)
{
	CONTR_DDR1 |= (1<<PIN_DET1) | (1<<PIN_DET2) | (1<<PIN_ATT) | (1<<PIN_AGC_ON) | (1<<PIN_AGC_LONG) | (1<<PIN_AGC_SHORT);
	CONTR_DDR2 |= (1<<PIN_BWC1) | (1<<PIN_BWC2);
	CONTR_DDR3 |= (1<<PIN_COR);
	CONTR_DDR_AUX |= (1<<PIN_STATUSLED);
}

void contr_process(void)																	// Processes dependencies in instr_state
{
	if( instr_error != 0 )																	// Is an error flag set?
	{
		instr_state |= (1UL<<MUTE);															// Mute receiver
		instr_state &= ~( (1UL<<COR_ACTIVE) );												// COR Active off
	}
	else if( (cor_level >= MIN_COR_LEVEL) || (instr_state & (1UL<<MGC)) )					// Is COR level set to min COR level or receiver is in MGC mode?
	{
		instr_state |= (1UL<<COR_ACTIVE);													// COR Active on
		if( !(instr_state & (1UL<<MUTE_SELECTED)) )											// Is Mute not selected on front panel?
			instr_state &= ~( (1UL<<MUTE) );												// Mute off
		else																				// Mute is selected on front panel
			instr_state |= (1UL<<MUTE);														// Mute receiver
	}
	else if( cor_level >= rssi_reading )													// Is measured RSSI level above COR level?
	{
		instr_state |= (1UL<<COR_ACTIVE);													// COR Active on
		if( !(instr_state & (1UL<<MUTE_SELECTED)) )											// Is Mute not selected on front panel?
			instr_state &= ~( (1UL<<MUTE) );												// Mute off
		else																				// Mute is selected on front panel
			instr_state |= (1UL<<MUTE);														// Mute receiver	
	}
	else
	{
		instr_state |= (1UL<<MUTE);															// Mute receiver
		instr_state &= ~( (1UL<<COR_ACTIVE) );												// COR Active off
	}
}

void contr_out(void)																		// Outputs instr_state to the MC ports
{
	static uint32_t instr_state_old;
	cli();
	uint32_t instr_state_loc = instr_state;													// Make local copy of instr_state
	sei();
	
	if( instr_state_loc != instr_state_old )												// Check if instr_state has changed
	{
		instr_state_old = instr_state_loc;
		// Output COR
		if( instr_state_loc & (1UL<<COR_ACTIVE) )
			CONTR_PORT3 |= (1<<PIN_COR);
		else
			CONTR_PORT3 &= ~( (1<<PIN_COR) );			
		// Process mute and detection mode
		if( instr_state_loc & (1UL<<MUTE) )
			CONTR_PORT1 |= (1<<PIN_DET1) | (1<<PIN_DET2);
		else if( instr_state_loc & (1UL<<AM) )
			CONTR_PORT1 &= ~( (1<<PIN_DET1) | (1<<PIN_DET2) );
		else if( instr_state_loc & (1UL<<USB) )
			CONTR_PORT1 &= ~(1<<PIN_DET1), CONTR_PORT1 |= (1<<PIN_DET2);
		else if( instr_state_loc & (1UL<<LSB) )
			CONTR_PORT1 |= (1<<PIN_DET1), CONTR_PORT1 &= ~( (1<<PIN_DET2) );
		// Process bandwidth select
		if( instr_state_loc & (1UL<<BW1) )
			CONTR_PORT2 |= (1<<PIN_BWC2), CONTR_PORT2 &= ~(1<<PIN_BWC1);
		else if( instr_state_loc & (1UL<<BW2) )
			CONTR_PORT2 |= (1<<PIN_BWC1), CONTR_PORT2 &= ~(1<<PIN_BWC2);
		else if( instr_state_loc & (1UL<<BW3) )
			CONTR_PORT2 |= (1<<PIN_BWC1) | (1<<PIN_BWC2);
		// Process attenuator
		if( instr_state_loc & (1UL<<ATT_ON) )
			CONTR_PORT1 |= (1<<PIN_ATT);
		else
			CONTR_PORT1 &= ~(1<<PIN_ATT);
		// Process AGC attack and dwell times
		if( instr_state_loc & (1UL<<AGC_FAST) )
			CONTR_PORT1 |= (1<<PIN_AGC_ON), CONTR_PORT1 |= (1<<PIN_AGC_SHORT), CONTR_PORT1 &= ~( (1<<PIN_AGC_LONG) );
		else if( instr_state_loc & (1UL<<AGC_SLOW) )
		{
			CONTR_PORT1 |= ( (1<<PIN_AGC_ON) | (1<<PIN_AGC_LONG) | (1<<PIN_AGC_SHORT) );	// Set fast and slow AGC for 250 ms after Slow was selected to allow AGC control loop to reach steady state quickly
			_delay_ms(250);
			CONTR_PORT1 |= (1<<PIN_AGC_ON), CONTR_PORT1 |= (1<<PIN_AGC_LONG), CONTR_PORT1 &= ~( (1<<PIN_AGC_SHORT) );
		}
		else if( instr_state_loc & (1UL<<AGC_HANG) )
		{
			CONTR_PORT1 |= ( (1<<PIN_AGC_ON) | (1<<PIN_AGC_LONG) | (1<<PIN_AGC_SHORT) );	// Set fast and slow AGC for 250 ms after Hang was selected to allow AGC control loop to  reach steady state quickly
			_delay_ms(250);
			CONTR_PORT1 |= (1<<PIN_AGC_ON), CONTR_PORT1 &= ~( (1<<PIN_AGC_LONG) ), CONTR_PORT1 &= ~( (1<<PIN_AGC_SHORT) );
		}
		// Process AGC or MGC mode
		if( instr_state_loc & (1<<MGC) )
			CONTR_PORT1 &= ~( (1<<PIN_AGC_ON) );
		else
			CONTR_PORT1 |= ( (1<<PIN_AGC_ON) );
	}
}

void status_LED_out(void)																	// Toggles the status LED on A3 board after a predetermined number of calls
{
	static uint16_t status_count;
	if( status_count < 1024 )
	{
		++status_count;
	}
	else
	{
		status_count = 0;
		CONTR_PORT_AUX ^= (1<<PIN_STATUSLED);
	}
}