/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * keyboard.c
 *
 * Version 1.0
 *
 * Created: 29.05.2017
 * Author: Mario Hellmich <mario.hellmich@web.de>
 *
 * Keyboard poll and updating of instr_state
 */ 


#include "def.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "keyboard.h"
#include "fault.h"

extern volatile uint32_t instr_state;
extern volatile uint8_t  instr_error;
extern volatile uint8_t  last_key;




/*
************************************************************************************************************************************************
Keyboard Poll:
The keyboard is polled every 10ms in the ISR corresponding to Timer2 Compare Match. If a key is found depressed, key_scan returns the
appropriate key number. The function key_scan is called by the ISR, and its result is stored in last_key. This variable is read by read_keyboard, 
which also clears it.
*/

#define ROWPORT					PORTC
#define ROWPIN					PINC
#define ROWDDR					DDRC
#define COLPORT					PORTD
#define COLPIN					PIND
#define COLDDR					DDRD
#define ROW1					PC0
#define ROW2					PC1
#define ROW3					PC2
#define ROW4					PC3
#define ROW5					PC4
#define COL1					PD4
#define COL2					PD5
#define COL3					PD6
#define COL4					PD7

void key_init(void)
{
	ROWDDR  |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Set rows as output
	COLDDR  &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );							// Set cols as input
	ROWPORT &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Rows low
	COLPORT |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );							// Cols pull-up on

	TCCR2 |= (1<<CS20) | (1<<CS22);															// Timer2 prescaler set to divide by 1024
	TCCR2 |= (1<<WGM21);																	// Clear Timer2 on compare
	TIMSK |= (1<<OCIE2);																	// Timer2 output compare match interrupt enable
	OCR2 = 156 - 1;																			// Compare register, set to 155 = 156-1, interrupt approx every 10ms
}

uint8_t key_scan(void)																		// Returns key number (keys are numbered from 1 to 20), and 0 for no key pressed, should be called every 10ms
{
	static uint8_t scan_dir;
	static uint8_t row;
	static uint8_t col;
	uint8_t key = 0;
	
	static uint8_t oldrow;
	static uint8_t oldcol;
	static uint8_t debounce;

	if( scan_dir == 0 )
	{
		if( ~COLPIN & (1 << COL1) )															// Read col
			col = 4;
		if( ~COLPIN & (1 << COL2) )
			col = 3;
		if( ~COLPIN & (1 << COL3) )
			col = 2;
		if( ~COLPIN & (1 << COL4) )
			col = 1;		
		
		ROWDDR  &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Set rows as input for scan of rows
		COLDDR  |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );						// Set cols as output for scan of rows
		COLPORT &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );						// Cols low for scan of rows
		ROWPORT |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Rows pull-up on for scan of rows
			
		scan_dir = 1;																		// Toggle scan direction; scan_dir=0 means scan cols, scan_dir=1 means scan rows
		return 0;
	}
	else
	{
		if( ~ROWPIN & (1<<ROW1) )															// Read row
			row = 5;
		if( ~ROWPIN & (1<<ROW2) )
			row = 9;
		if( ~ROWPIN & (1<<ROW3) )
			row = 13;
		if( ~ROWPIN & (1<<ROW4) )
			row = 17;
		if( ~ROWPIN & (1<<ROW5) )
			row = 21;
			
		ROWDDR  |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Set rows as output for scan of columns
		COLDDR  &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );						// Set cols as input for scan of columns
		ROWPORT &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Rows low for scan of columns
		COLPORT |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );						// Cols pull-up on for scan of columns
					
		scan_dir = 0;																		// Toggle scan direction
		key = row - col;																	// Determine key number, yields zero if no key is pressed
		
		if( oldrow != row  || oldcol != col )												// Debouncing algorithm
		{
			oldrow = row;
			oldcol = col;
			debounce = 2;
		}		
		row = 0;
		col = 0;		
		
		if( key != 0 && debounce != 0 && --debounce == 0 )									// Old key equals key for four calls (= 40 ms)
			return key;
		return 0;	
	}
}

ISR(TIMER2_COMP_vect)																		// Interrupt service routine for Timer2 compare match
{
	uint8_t current_key = key_scan();
	if( current_key != 0 && last_key == 0 )													// Ignore further key presses if last_key has not been processed and cleared by read_keyboard
	{
		last_key = current_key;
	}
}




/*
************************************************************************************************************************************************
Read Data from last_key and Update instr_state Appropriately
*/

void read_keyboard(void)																	// Updates instr_state from last_key
{
	uint8_t last_key_loc;
	cli();
	last_key_loc = last_key;																// Make local copy of last_key
	last_key = 0;																			// Clear last_key
	sei();
	if( (last_key_loc != 0) && (instr_error == 0) )											// Any key pressed and no error flag set?
	{
		switch( last_key_loc )																// If yes, take appropriate action
		{
			// Interlocked switches
			case KEY_AM:
				instr_state |= (1UL<<AM), instr_state &= ~((1UL<<USB) | (1UL<<LSB));
					break;
			case KEY_USB:
				instr_state |= (1UL<<USB), instr_state &= ~((1UL<<AM) | (1UL<<LSB));
					break;
			case KEY_LSB:
				instr_state |= (1UL<<LSB), instr_state &= ~((1<<AM) | (1<<USB));
					break;
			case KEY_FAST:
				instr_state |= (1UL<<AGC_FAST), instr_state &= ~((1UL<<AGC_HANG) | (1UL<<AGC_SLOW) | (1UL<<MGC));
					break;
			case KEY_SLOW:
				instr_state |= (1UL<<AGC_SLOW), instr_state &= ~((1UL<<AGC_HANG) | (1UL<<AGC_FAST) | (1UL<<MGC));
					break;
			case KEY_HANG:
				instr_state |= (1UL<<AGC_HANG), instr_state &= ~((1UL<<AGC_FAST) | (1UL<<AGC_SLOW) | (1UL<<MGC));
					break;
			case KEY_MGC:
				instr_state |= (1UL<<MGC), instr_state &= ~((1UL<<AGC_FAST) | (1UL<<AGC_SLOW) | (1UL<<AGC_HANG));
					break;				
			case KEY_BW1:
				instr_state |= (1UL<<BW1), instr_state &= ~((1UL<<BW2) | (1UL<<BW3));
					break;
			case KEY_BW2:
				instr_state |= (1UL<<BW2), instr_state &= ~((1UL<<BW1) | (1UL<<BW3));
					break;
			case KEY_BW3:
				instr_state |= (1UL<<BW3), instr_state &= ~((1UL<<BW2) | (1UL<<BW1));
					break;
			case KEY_100KHZ:
				instr_state |= (1UL<<TUNE_100KHZ), instr_state &= ~((1UL<<TUNE_10KHZ) | (1UL<<TUNE_1KHZ) | (1UL<<TUNE_100HZ) | (1UL<<TUNE_10HZ) | (1UL<<TUNE_LOCK) | (1UL<<SET_COR) | (1UL<<BFO_TUNE));
					break;
			case KEY_10KHZ:
				instr_state |= (1UL<<TUNE_10KHZ), instr_state &= ~((1UL<<TUNE_100KHZ) | (1UL<<TUNE_1KHZ) | (1UL<<TUNE_100HZ) | (1UL<<TUNE_10HZ) | (1UL<<TUNE_LOCK) | (1UL<<SET_COR) | (1UL<<BFO_TUNE));
					break;
			case KEY_1KHZ:
				instr_state |= (1UL<<TUNE_1KHZ), instr_state &= ~((1UL<<TUNE_100KHZ) | (1UL<<TUNE_10KHZ) | (1UL<<TUNE_100HZ) | (1UL<<TUNE_10HZ) | (1UL<<TUNE_LOCK) | (1UL<<SET_COR) | (1UL<<BFO_TUNE));
					break;
			case KEY_100HZ:
				instr_state |= (1UL<<TUNE_100HZ), instr_state &= ~((1UL<<TUNE_100KHZ) | (1UL<<TUNE_10KHZ) | (1UL<<TUNE_1KHZ) | (1UL<<TUNE_10HZ) | (1UL<<TUNE_LOCK) | (1UL<<SET_COR) | (1UL<<BFO_TUNE));
					break;
			case KEY_10HZ:
				instr_state |= (1UL<<TUNE_10HZ), instr_state &= ~((1UL<<TUNE_100KHZ) | (1UL<<TUNE_10KHZ) | (1UL<<TUNE_1KHZ) | (1UL<<TUNE_100HZ) | (1UL<<TUNE_LOCK) | (1UL<<SET_COR) | (1UL<<BFO_TUNE));
					break;
			case KEY_TLOCK:
				instr_state |= (1UL<<TUNE_LOCK), instr_state &= ~((1UL<<TUNE_100KHZ) | (1UL<<TUNE_10KHZ) | (1UL<<TUNE_1KHZ) | (1UL<<TUNE_100HZ) | (1UL<<TUNE_10HZ) | (1UL<<SET_COR) | (1UL<<BFO_TUNE));				
					break;
			// Toggle switches			
			case KEY_MUTE:
				instr_state ^= (1UL<<MUTE_SELECTED);
					break;
			case KEY_ATT:
				instr_state ^= (1UL<<ATT_ON);
					break;				
			// Toggle switches with interlock
			case KEY_SETCOR:
				instr_state ^= (1UL<<SET_COR), instr_state &= ~((1UL<<BFO_TUNE));
					break;
			case KEY_BFOTUNE:
				instr_state ^= (1UL<<BFO_TUNE), instr_state &= ~((1UL<<SET_COR));
					break;
		}
	}
}




/*
************************************************************************************************************************************************
Test for a stuck key at startup
*/

void stuck_key_test(void)
{
	ROWDDR  &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Set rows as input for scan of rows
	COLDDR  |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );							// Set cols as output for scan of rows
	COLPORT &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) );							// Cols low for scan of rows
	ROWPORT |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Rows pull-up on for scan of rows	
	_delay_ms(20);
	if( ~ROWPIN & ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) ) )			// Any row not high?
		handle_error(ERR_STUCK_KEY);
}