/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this 
 * notice you can do whatever you want with this stuff. If we meet some day, 
 * and you think this stuff is worth it, you can buy me a beer in return. 
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * main.c
 *
 * Version 1.0
 *
 * Created: 29.05.2017
 * Author: Mario Hellmich <mario.hellmich@web.de>
 *
 * This is the firmware for the HF Receiver.
 *
 * Target Device:	ATmega64A
 * CPU Clock:		16 MHz
 * VDD:				5 V
 *
 */ 


#include "def.h"																			// Preprocessor directives for instr_state bits and key codes, CPU clock, and global variables

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "annunciators.h"																	// Annunciator multiplexing and output
#include "keyboard.h"																		// Keyboard poll and updating of instr_state
#include "encoder.h"																		// Rotary encoder poll and updating of appropriate globals
#include "ad.h"																				// A/D converter control and processing
#include "instr_control.h"																	// Data output to the MCU ports
#include "si5341.h"																			// Controls the Si5341 PLL synthesizer for LO1, LO2, and the BFO, via the TWI bus
#include "fault.h"																			// Self check functions and hardware error handling




/*
************************************************************************************************************************************************
Global Variables for Instrument State and Data Exchange with the ISR
*/

volatile uint32_t instr_state;			// Basic instrument state, the roles of the individual bits are defined in def.h
volatile uint8_t  instr_error;			// Instrument error status, the roles of the individual bits are defined in def.h
volatile uint8_t  annunc_pattern[13];	// LED pattern on seven segment displays and annunciators that is to be displayed in current mode and with current frequency and RSSI value
volatile uint32_t tuned_freq;			// Frequency the receiver is tuned to in Hz, this number is displayed and outputted to LO1, range: 00.001,000 MHz to 30.000,000 MHz
volatile int16_t  bfo_offset;			// Current BFO offset frequency, it is displayed in BFO-Tune Mode and outputted to BFO, range -3,500 to +3,500 Hz
volatile uint8_t  cor_level;			// Current COR level in -dBm 
volatile uint8_t  rssi_reading;			// Current RSSI display reading as measured by the A/D converter in -dBm or in percent demodulator usage when receiver is in MGC mode
volatile uint8_t  last_key;				// Contains the code of the last key pressed, and is reset to zero when key press is processed




/*
************************************************************************************************************************************************
Prototypes for functions defined below in this file
*/
void annunc_test(void);
void annunc_off(void);
void instr_init(void);
void instr_start_state(void);




/*
************************************************************************************************************************************************
The Main Function
*/

int main(void)
{
	instr_init();									// Initialize instr_state, mute audio output
	annunc_init();									// Initialize annunciator multiplex
	contr_init();									// Initialize instrument control outputs
		contr_out();								// Output instr_state once to mute receiver during startup
	TWI_init();										// Initialize TWI bus
	fault_init();									// Initialize HW self check functions	
		stuck_key_test();							// Test for stuck keys
	key_init();										// Initialize keyboard polling	
	
	sei();											// Global interrupt enable
	
	annunc_test();									// Annunciator test, light up all annunciators
	si5341_init();									// Initialize Si5341 frequency synthesizer
	freq_init();									// Set initial frequencies for LO1, LO2, and BFO	
	
	_delay_ms(2000);								// Wait two seconds for annunciator test, then proceed	
	
	cli();
	enc_init();										// Initialize rotary encoder
	adc_init();										// Initialize A/D converter
	sei();
		
	annunc_off();									// Annunciators off for 0.5 sec
	_delay_ms(500);	
	instr_start_state();							// Set instr_state to default power on state
	
	wdt_enable(WDTO_500MS);							// Enable watchdog timer and set timeout to 0.5 seconds
	
	
    while(1)										// Main Loop
    {
	PLL_fault_test();								// Periodic HW self check		
	read_keyboard();								// Read keyboard state and update instr_state, reset last_key to 0
	read_enc();										// Read rotary encoder and update appropriate variables
	disp_out();										// Output numerical display to annunc_pattern
	annunc_out();									// Output annunciators to annunc_pattern
	status_LED_out();								// Output status LED on control board A3
	
	contr_process();								// Process dependencies in instr_state
	adc_process();									// Processes ADC result and write to appropriate global
	contr_out();									// Output instr_state to the MC ports
	
	freq_update();									// Update the LO1 and BFO frequencies
	
	wdt_reset();									// Reset watchdog timer
    }
	return 0;
}




/*
************************************************************************************************************************************************
Start Condition of Globals, Display and Annunciator Test:
*/

void annunc_test(void)
{
	annunc_pattern [0]=  0xFF;
	annunc_pattern [1]=  0xFF;
	annunc_pattern [2]=  0xFF;
	annunc_pattern [3]=  0xFF;
	annunc_pattern [4]=  0xFF;
	annunc_pattern [5]=  0xFF;
	annunc_pattern [6]=  0xFF;
	annunc_pattern [7]=  0xFF;
	annunc_pattern [8]=  0xFF;
	annunc_pattern [9]=  0xFF;
	annunc_pattern [10]= 0xFF;
	annunc_pattern [11]= 0xFF;
	annunc_pattern [12]= 0xFF;
}

void annunc_off(void)
{
	annunc_pattern [0]=  0x00;
	annunc_pattern [1]=  0x00;
	annunc_pattern [2]=  0x00;
	annunc_pattern [3]=  0x00;
	annunc_pattern [4]=  0x00;
	annunc_pattern [5]=  0x00;
	annunc_pattern [6]=  0x00;
	annunc_pattern [7]=  0x00;
	annunc_pattern [8]=  0x00;
	annunc_pattern [9]=  0x00;
	annunc_pattern [10]= 0x00;
	annunc_pattern [11]= 0x00;
	annunc_pattern [12]= 0x00;
}

void instr_init(void)
{
	instr_state = (1UL<<AM) | (1UL<<MUTE) | (1UL<<MUTE_SELECTED) | (1UL<<AGC_FAST) | (1<<BW1) | (1UL<<TUNE_LOCK) | (1UL<<COR_ACTIVE);
	instr_error = 0;
	tuned_freq  = 7200000;
	bfo_offset  = 0;
	cor_level   = MIN_COR_LEVEL;
	last_key = 0;
}

void instr_start_state(void)
{
	instr_state = ( (1UL<<AM) | (1UL<<AGC_FAST) | (1<<BW1) | (1UL<<TUNE_LOCK) | (1UL<<COR_ACTIVE) | (1UL<<ATT_ON) ) & ~( (1UL<<MUTE) | (1UL<<MUTE_SELECTED) );
	tuned_freq  = 7200000;
	bfo_offset  = 0;
	cor_level   = MIN_COR_LEVEL;
	last_key = 0;
}